/*
 * IPVS:        Locality-Based Least-Connection scheduling module
 *
 * Version:     $Id: ip_vs_lblc.c,v 1.9 2002/03/25 12:44:35 wensong Exp $
 *
 * Authors:     Wensong Zhang <wensong@gnuchina.org>
 *
 *              This program is free software; you can redistribute it and/or
 *              modify it under the terms of the GNU General Public License
 *              as published by the Free Software Foundation; either version
 *              2 of the License, or (at your option) any later version.
 *
 * Changes:
 *     Martin Hamilton         :    fixed the terrible locking bugs
 *                                   *lock(tbl->lock) ==> *lock(&tbl->lock)
 *     Wensong Zhang           :    fixed the uninitilized tbl->lock bug
 *     Wensong Zhang           :    added doing full expiration check to
 *                                   collect stale entries of 24+ hours when
 *                                   no partial expire check in a half hour
 *     Julian Anastasov        :    replaced del_timer call with del_timer_sync
 *                                   to avoid the possible race between timer
 *                                   handler and del_timer thread in SMP
 *
 */

/*
  *	Port IPVS to FreeBSD
  *	Authors:		Li Wang <dragonfly@linux-vs.org>
  *
  */
  
/*
 * The lblc algorithm is as follows (pseudo code):
 *
 *       if cachenode[dest_ip] is null then
 *               n, cachenode[dest_ip] <- {weighted least-conn node};
 *       else
 *               n <- cachenode[dest_ip];
 *               if (n is dead) OR
 *                  (n.conns>n.weight AND
 *                   there is a node m with m.conns<m.weight/2) then
 *                 n, cachenode[dest_ip] <- {weighted least-conn node};
 *
 *       return n;
 *
 * Thanks must go to Wenzhuo Zhang for talking WCCP to me and pushing
 * me to write this module.
 */


#include "../../ip_vs.h"


/*
 *    It is for garbage collection of stale IPVS lblc entries,
 *    when the table is full.
 */
#define CHECK_EXPIRE_INTERVAL   	(60*HZ)
#define ENTRY_TIMEOUT           		(6*60*HZ)

/*
 *    It is for full expiration check.
 *    When there is no partial expiration check (garbage collection)
 *    in a half hour, do a full expiration check to collect stale
 *    entries that haven't been touched for a day.
 */
#define COUNT_FOR_FULL_EXPIRATION   30
int sysctl_ip_vs_lblc_expiration = 24*60*60*100;


/*
 *     for IPVS lblc entry hash table
 */
#ifndef CONFIG_IP_VS_LBLC_TAB_BITS
#define CONFIG_IP_VS_LBLC_TAB_BITS      10
#endif
#define IP_VS_LBLC_TAB_BITS     CONFIG_IP_VS_LBLC_TAB_BITS
#define IP_VS_LBLC_TAB_SIZE     (1 << IP_VS_LBLC_TAB_BITS)
#define IP_VS_LBLC_TAB_MASK     (IP_VS_LBLC_TAB_SIZE - 1)


/*
 *      IPVS lblc entry represents an association between destination
 *      IP address and its destination server
 */
struct ip_vs_lblc_entry {
	LIST_ENTRY(ip_vs_lblc_entry)		list;
	u_int32_t                   addr;           /* destination IP address */
	struct ip_vs_dest       *dest;          /* real server (cache) */
	unsigned long           lastuse;        /* last used time */
};


/*
 *      IPVS lblc hash table
 */
struct ip_vs_lblc_table {
	struct sx		 lock;			/* lock for this table */
	LIST_HEAD(ip_vs_lblc_entry_list, ip_vs_lblc_entry)	bucket[IP_VS_LBLC_TAB_SIZE];  /* hash bucket */
	volatile u_int32_t                entries;        /* number of entries */
	int                     max_size;       /* maximum size of entries */
	struct callout		   periodic_timer; /* collect stale entries */
	int                     rover;          /* rover for expire check */
	int                     counter;        /* counter for no expire */
};

/*
 *      new/free a ip_vs_lblc_entry, which is a mapping of a destionation
 *      IP address to a server.
 */
static inline struct ip_vs_lblc_entry *
ip_vs_lblc_new(u_int32_t daddr, struct ip_vs_dest *dest)
{
	struct ip_vs_lblc_entry *en;

	en = malloc(sizeof(struct ip_vs_lblc_entry), M_IPVSBUF, M_NOWAIT|M_ZERO);
	if (en == NULL) {
		IP_VS_ERR("ip_vs_lblc_new(): no memory\n");
		return NULL;
	}

	en->addr = daddr;

	atomic_add_int(&dest->refcnt, 1);
	en->dest = dest;

	return en;
}


static inline void ip_vs_lblc_free(struct ip_vs_lblc_entry *en)
{
	LIST_REMOVE(en, list);
	/*
	 * We don't kfree dest because it is refered either by its service
	 * or the trash dest list.
	 */
	atomic_subtract_int(&en->dest->refcnt, 1);
	free(en, M_IPVSBUF);
}


/*
 *	Returns hash value for IPVS LBLC entry
 */
static inline unsigned ip_vs_lblc_hashkey(u_int32_t addr)
{
	return (ntohl(addr)*2654435761UL) & IP_VS_LBLC_TAB_MASK;
}


/*
 *	Hash an entry in the ip_vs_lblc_table.
 *	returns bool success.
 */
static int
ip_vs_lblc_hash(struct ip_vs_lblc_table *tbl, struct ip_vs_lblc_entry *en)
{
	unsigned hash;

	if (ENTRY_ON_LIST(en, list)) {
		IP_VS_ERR("ip_vs_lblc_hash(): request for already hashed, "
			  "called from %p\n", __builtin_return_address(0));
		return 0;
	}

	/*
	 *	Hash by destination IP address
	 */
	hash = ip_vs_lblc_hashkey(en->addr);

	sx_xlock(&tbl->lock);
	LIST_INSERT_HEAD(&tbl->bucket[hash], en, list);
	atomic_add_int(&tbl->entries, 1);
	sx_xunlock(&tbl->lock);

	return 1;
}


#if 0000
/*
 *	Unhash ip_vs_lblc_entry from ip_vs_lblc_table.
 *	returns bool success.
 */
static int ip_vs_lblc_unhash(struct ip_vs_lblc_table *tbl,
			     struct ip_vs_lblc_entry *en)
{
	if (list_empty(&en->list)) {
		IP_VS_ERR("ip_vs_lblc_unhash(): request for not hashed entry, "
			  "called from %p\n", __builtin_return_address(0));
		return 0;
	}

	/*
	 * Remove it from the table
	 */
	write_lock(&tbl->lock);
	list_del(&en->list);
	INIT_LIST_HEAD(&en->list);
	write_unlock(&tbl->lock);

	return 1;
}
#endif


/*
 *  Get ip_vs_lblc_entry associated with supplied parameters.
 */
static inline struct ip_vs_lblc_entry *
ip_vs_lblc_get(struct ip_vs_lblc_table *tbl, u_int32_t addr)
{
	unsigned hash;
	struct ip_vs_lblc_entry *en;
	struct ip_vs_lblc_entry_list *l;

	hash = ip_vs_lblc_hashkey(addr);
	l = &tbl->bucket[hash];

	sx_slock(&tbl->lock);

	LIST_FOREACH(en, l, list) {
		if (en->addr == addr) {
			/* HIT */
			sx_sunlock(&tbl->lock);
			return en;
		}
	}

	sx_sunlock(&tbl->lock);

	return NULL;
}


/*
 *      Flush all the entries of the specified table.
 */
static void ip_vs_lblc_flush(struct ip_vs_lblc_table *tbl)
{
	int i;
	struct ip_vs_lblc_entry *en, *tmp;

	for (i=0; i<IP_VS_LBLC_TAB_SIZE; i++) {
		sx_xlock(&tbl->lock);
		LIST_FOREACH_SAFE(en, &tbl->bucket[i], list, tmp) {
			ip_vs_lblc_free(en);
			atomic_subtract_int(&tbl->entries, 1);
		}
		sx_xunlock(&tbl->lock);
	}
}


static inline void ip_vs_lblc_full_check(struct ip_vs_lblc_table *tbl)
{
	unsigned long now = ticks;
	int i, j;
	struct ip_vs_lblc_entry_list *l;
	struct ip_vs_lblc_entry *en, *tmp;

	for (i=0, j=tbl->rover; i<IP_VS_LBLC_TAB_SIZE; i++) {
		j = (j + 1) & IP_VS_LBLC_TAB_MASK;
		l = &tbl->bucket[j];
		sx_xlock(&tbl->lock);
		LIST_FOREACH_SAFE(en, l, list, tmp) {
			if ((now - en->lastuse) <
			    sysctl_ip_vs_lblc_expiration) {
				continue;
			}
			ip_vs_lblc_free(en);
			atomic_subtract_int(&tbl->entries, 1);
		}
		
		sx_xunlock(&tbl->lock);
	}
	tbl->rover = j;
}


/*
 *      Periodical timer handler for IPVS lblc table
 *      It is used to collect stale entries when the number of entries
 *      exceeds the maximum size of the table.
 *
 *      Fixme: we probably need more complicated algorithm to collect
 *             entries that have not been used for a long time even
 *             if the number of entries doesn't exceed the maximum size
 *             of the table.
 *      The full expiration check is for this purpose now.
 */
static void ip_vs_lblc_check_expire(void *data)
{
	struct ip_vs_lblc_table *tbl;
	unsigned long now = ticks;
	int goal;
	int i, j;
	struct ip_vs_lblc_entry_list *l;
	struct ip_vs_lblc_entry *en, *tmp;

	tbl = (struct ip_vs_lblc_table *)data;
	if ((tbl->counter % COUNT_FOR_FULL_EXPIRATION) == 0) {
		/* do full expiration check */
		ip_vs_lblc_full_check(tbl);
		tbl->counter = 1;
		goto out;
	}

	if (atomic_read(&tbl->entries) <= tbl->max_size) {
		tbl->counter++;
		goto out;
	}
	goal = (atomic_read(&tbl->entries) - tbl->max_size)*4/3;
	if (goal > tbl->max_size/2)
		goal = tbl->max_size/2;

	for (i=0, j=tbl->rover; i<IP_VS_LBLC_TAB_SIZE; i++) {
		j = (j + 1) & IP_VS_LBLC_TAB_MASK;
		l = &tbl->bucket[j];
		sx_xlock(&tbl->lock);
		LIST_FOREACH_SAFE(en, l, list, tmp) {
			if ((now - en->lastuse) < ENTRY_TIMEOUT) {
				continue;
			}
			ip_vs_lblc_free(en);
			atomic_subtract_int(&tbl->entries, 1);
			goal--;

		}
		
		sx_xunlock(&tbl->lock);
		if (goal <= 0)
			break;
	}
	tbl->rover = j;

  out:
  	callout_reset(&tbl->periodic_timer, CHECK_EXPIRE_INTERVAL, ip_vs_lblc_check_expire, tbl);
}


static int ip_vs_lblc_init_svc(struct ip_vs_service *svc)
{
	int i;
	struct ip_vs_lblc_table *tbl;

	/*
	 *    Allocate the ip_vs_lblc_table for this service
	 */
	tbl = malloc(sizeof(struct ip_vs_lblc_table), M_IPVSBUF, M_NOWAIT|M_ZERO);
	if (tbl == NULL) {
		IP_VS_ERR("ip_vs_lblc_init_svc(): no memory\n");
		return ENOMEM;
	}
	svc->sched_data = tbl;
	IP_VS_DBG(6, "LBLC hash table (memory=%dbytes) allocated for "
		  "current service\n",
		  sizeof(struct ip_vs_lblc_table));

	/*
	 *    Initialize the hash buckets
	 */
	for (i=0; i<IP_VS_LBLC_TAB_SIZE; i++) {
		LIST_INIT(&tbl->bucket[i]);
	}
	sx_init(&tbl->lock, "IPVS lblc table lock");
	tbl->max_size = IP_VS_LBLC_TAB_SIZE*16;
	tbl->rover = 0;
	tbl->counter = 1;

	/*
	 *    Hook periodic timer for garbage collection
	 */
	callout_init(&tbl->periodic_timer, 1);
	callout_reset(&tbl->periodic_timer, CHECK_EXPIRE_INTERVAL, ip_vs_lblc_check_expire, tbl);
	
	return 0;
}


static int ip_vs_lblc_done_svc(struct ip_vs_service *svc)
{
	struct ip_vs_lblc_table *tbl = svc->sched_data;

	/* remove periodic timer */
	callout_drain(&tbl->periodic_timer);

	/* got to clean up table entries here */
	ip_vs_lblc_flush(tbl);

	/* release the table itself */
	free(svc->sched_data, M_IPVSBUF);
	IP_VS_DBG(6, "LBLC hash table (memory=%dbytes) released\n",
		  sizeof(struct ip_vs_lblc_table));

	return 0;
}


static int ip_vs_lblc_update_svc(struct ip_vs_service *svc)
{
	return 0;
}


static inline struct ip_vs_dest *
__ip_vs_wlc_schedule(struct ip_vs_service *svc, struct ip *iph)
{
	register struct ip_vs_dest_list *l;
	struct ip_vs_dest *dest, *least;
	int loh, doh;

	/*
	 * We think the overhead of processing active connections is fifty
	 * times higher than that of inactive connections in average. (This
	 * fifty times might not be accurate, we will change it later.) We
	 * use the following formula to estimate the overhead:
	 *                dest->activeconns*50 + dest->inactconns
	 * and the load:
	 *                (dest overhead) / dest->weight
	 *
	 * Remember -- no floats in kernel mode!!!
	 * The comparison of h1*w2 > h2*w1 is equivalent to that of
	 *                h1/w1 > h2/w2
	 * if every weight is larger than zero.
	 *
	 * The server with weight=0 is quiesced and will not receive any
	 * new connection.
	 */

	l = &svc->destinations;
	LIST_FOREACH(least, l, n_list) {
		if (atomic_read(&least->weight) > 0) {
			loh = atomic_read(&least->activeconns) * 50
				+ atomic_read(&least->inactconns);
			goto nextstage;
		}
	}
	
	return NULL;

	/*
	 *    Find the destination with the least load.
	 */
  nextstage:
  	for (dest = LIST_NEXT(least, n_list); dest; dest = LIST_NEXT(dest, n_list)) {
		doh = atomic_read(&dest->activeconns) * 50
			+ atomic_read(&dest->inactconns);
		if (loh * atomic_read(&dest->weight) >
		    doh * atomic_read(&least->weight)) {
			least = dest;
			loh = doh;
		}
  	}
	
	IP_VS_DBG(6, "LBLC: server %d.%d.%d.%d:%d "
		  "activeconns %d refcnt %d weight %d overhead %d\n",
		  NIPQUAD(least->addr), ntohs(least->port),
		  atomic_read(&least->activeconns),
		  atomic_read(&least->refcnt),
		  atomic_read(&least->weight), loh);

	return least;
}


/*
 *   If this destination server is overloaded and there is a less loaded
 *   server, then return true.
 */
static inline int
is_overloaded(struct ip_vs_dest *dest, struct ip_vs_service *svc)
{
	if (atomic_read(&dest->activeconns) > atomic_read(&dest->weight)) {
		register struct ip_vs_dest_list *l;
		struct ip_vs_dest *d;

		l = &svc->destinations;
		LIST_FOREACH(d, l, n_list) {
			if (atomic_read(&d->activeconns)*2
			    < atomic_read(&d->weight)) {
				return 1;
			}
		}
	}
	return 0;
}


/*
 *    Locality-Based (weighted) Least-Connection scheduling
 */
static struct ip_vs_dest *
ip_vs_lblc_schedule(struct ip_vs_service *svc, struct ip *iph)
{
	struct ip_vs_dest *dest;
	struct ip_vs_lblc_table *tbl;
	struct ip_vs_lblc_entry *en;

	IP_VS_DBG(6, "ip_vs_lblc_schedule(): Scheduling...\n");

	tbl = (struct ip_vs_lblc_table *)svc->sched_data;
	en = ip_vs_lblc_get(tbl, iph->ip_dst.s_addr);
	if (en == NULL) {
		dest = __ip_vs_wlc_schedule(svc, iph);
		if (dest == NULL) {
			IP_VS_DBG(1, "no destination available\n");
			return NULL;
		}
		en = ip_vs_lblc_new(iph->ip_dst.s_addr, dest);
		if (en == NULL) {
			return NULL;
		}
		ip_vs_lblc_hash(tbl, en);
	} else {
		dest = en->dest;
		if (!(dest->flags & IP_VS_DEST_F_AVAILABLE)
		    || atomic_read(&dest->weight) <= 0
		    || is_overloaded(dest, svc)) {
			dest = __ip_vs_wlc_schedule(svc, iph);
			if (dest == NULL) {
				IP_VS_DBG(1, "no destination available\n");
				return NULL;
			}
			atomic_subtract_int(&en->dest->refcnt, 1);
			atomic_add_int(&dest->refcnt, 1);
			en->dest = dest;
		}
	}
	en->lastuse = ticks;

	IP_VS_DBG(6, "LBLC: destination IP address %u.%u.%u.%u "
		  "--> server %u.%u.%u.%u:%d\n",
		  NIPQUAD(en->addr),
		  NIPQUAD(dest->addr),
		  ntohs(dest->port));

	return dest;
}


/*
 *      IPVS LBLC Scheduler structure
 */
static struct ip_vs_scheduler ip_vs_lblc_scheduler =
{
	{0},                    /* n_list */
	"lblc",                 /* name */
	0,         					/* refcnt */
	NULL,					/* this module */
	ip_vs_lblc_init_svc,    /* service initializer */
	ip_vs_lblc_done_svc,    /* service done */
	ip_vs_lblc_update_svc,  /* service updater */
	ip_vs_lblc_schedule,    /* select a server from the destination list */
};


static int ip_vs_lblc_init(void)
{
	return register_ip_vs_scheduler(&ip_vs_lblc_scheduler);
}


static void ip_vs_lblc_cleanup(void)
{
	unregister_ip_vs_scheduler(&ip_vs_lblc_scheduler);
}

/*
 * Load handler that deals with the loading and unloading of a KLD.
 */
static int ipvs_lblc_loader(struct module *m, int what, void *arg)
{
 	int err = 0;
	
    	switch (what) {
    	case MOD_LOAD:
		err=ip_vs_lblc_init();
        	break;
    	case MOD_UNLOAD:
	case MOD_SHUTDOWN:
		ip_vs_lblc_cleanup();
        	break;
    	default:
		err = EOPNOTSUPP;
      		break;
    	}

    	return (err);
}

static moduledata_t ipvs_lblc_mod = {
    "ipvs_lblc",
    ipvs_lblc_loader,
    NULL
};

DECLARE_MODULE(ipvs_lblc, ipvs_lblc_mod,SI_SUB_KLD, SI_ORDER_ANY);
MODULE_DEPEND(ipvs_lblc, ipvs, 1, 1, 1);
