/*
 * IPVS:        Round-Robin Scheduling module
 *
 * Version:     $Id: ip_vs_rr.c,v 1.8 2001/10/19 15:05:17 wensong Exp $
 *
 * Authors:     Wensong Zhang <wensong@linuxvirtualserver.org>
 *              Peter Kese <peter.kese@ijs.si>
 *
 *              This program is free software; you can redistribute it and/or
 *              modify it under the terms of the GNU General Public License
 *              as published by the Free Software Foundation; either version
 *              2 of the License, or (at your option) any later version.
 *
 * Fixes/Changes:
 *     Wensong Zhang            :     changed the ip_vs_rr_schedule to return dest
 *     Julian Anastasov         :     fixed the NULL pointer access bug in debugging
 *     Wensong Zhang            :     changed some comestics things for debugging
 *     Wensong Zhang            :     changed for the d-linked destination list
 *     Wensong Zhang            :     added the ip_vs_rr_update_svc
 *     Wensong Zhang            :     added any dest with weight=0 is quiesced
 *
 */

/*
  *	Port IPVS to FreeBSD
  *	Authors:		Li Wang <dragonfly@linux-vs.org>
  *
  */

#include "../../ip_vs.h"

static int ip_vs_rr_init_svc(struct ip_vs_service *svc)
{
	svc->sched_data = &svc->destinations;
	return 0;
}


static int ip_vs_rr_done_svc(struct ip_vs_service *svc)
{
	return 0;
}


static int ip_vs_rr_update_svc(struct ip_vs_service *svc)
{
	svc->sched_data = &svc->destinations;
	return 0;
}


/*
 * Round-Robin Scheduling
 */
static struct ip_vs_dest *
ip_vs_rr_schedule(struct ip_vs_service *svc, struct ip *iph)
{
	register void *p, *q;	
	struct ip_vs_dest *dest = NULL;
	int hit = 0;
	
	IP_VS_DBG(6, "ip_vs_rr_schedule(): Scheduling...\n");

	sx_xlock(&svc->sched_lock);
	p = q = svc->sched_data;
	do {
		if (p == &svc->destinations) { 
			dest = LIST_FIRST(&svc->destinations);
			if (!dest) {
				sx_xunlock(&svc->sched_lock);
				return NULL;
			}
		} else {
			dest = LIST_NEXT((struct ip_vs_dest *)p, n_list);
			if (!dest) {
				p = &svc->destinations;
				continue;
			}
		}
		if (atomic_read(&dest->weight) > 0)
			hit = 1;
		else
			p = dest;

	} while ((!hit) && (p != q));

	
  if (hit)
	svc->sched_data = dest;
	sx_xunlock(&svc->sched_lock);
	IP_VS_DBG(6, "RR: server %u.%u.%u.%u:%u "
		  "activeconns %d refcnt %d weight %d\n",
		  NIPQUAD(dest->addr), ntohs(dest->port),
		  atomic_read(&dest->activeconns),
		  atomic_read(&dest->refcnt), atomic_read(&dest->weight));

	return dest;
}


static struct ip_vs_scheduler ip_vs_rr_scheduler = {
	{0},					/* n_list */
	"rr",					/* name */
	0,					/* refcnt */
	NULL,				/* this module */
	ip_vs_rr_init_svc,	/* service initializer */
	ip_vs_rr_done_svc,	/* service done */
	ip_vs_rr_update_svc,	/* service updater */
	ip_vs_rr_schedule,	/* select a server from the destination list */
};

static int ip_vs_rr_init(void)
{
	return register_ip_vs_scheduler(&ip_vs_rr_scheduler);
}

static void ip_vs_rr_cleanup(void)
{
	unregister_ip_vs_scheduler(&ip_vs_rr_scheduler);
}

/*
 * Load handler that deals with the loading and unloading of a KLD.
 */
static int ipvs_rr_loader(struct module *m, int what, void *arg)
{
 	int err = 0;
	
    	switch (what) {
    	case MOD_LOAD:
		err=ip_vs_rr_init();
        	break;
    	case MOD_UNLOAD:
	case MOD_SHUTDOWN:
		ip_vs_rr_cleanup();
        	break;
    	default:
		err = EOPNOTSUPP;
      		break;
    	}

    	return (err);
}

static moduledata_t ipvs_rr_mod = {
    "ipvs_rr",
    ipvs_rr_loader,
    NULL
};

DECLARE_MODULE(ipvs_rr, ipvs_rr_mod,SI_SUB_KLD, SI_ORDER_ANY);
MODULE_DEPEND(ipvs_rr, ipvs, 1, 1, 1);

